var _ = require('lodash');
var frp = require('frp');
var cell = require('frp').cell;
var R = require('mxd').Rect;
var V = require('mxd').Vector;
var assert = require('assert');

var T = require('ace/transform');
var makeView = require('ace/view');
var util = require('ace/util');
var canvasUtils = require('ace/canvasUtils');


exports.loadImage = function(url) {
  var r = cell();
  var img = new Image();
  img.src = url;
  img.onload = function() {
    r.set(img);
  };
  return r.signal;
};

function def(name, fn) {
  exports[name] = util.sigargs(fn);
}

/*
  N.B. functions defined with 'def' here will have their
  arguments converted to FRP signals, except arguments whose
  value is 'undefined'.
*/

def('rectangle', function(rect, style) {

  style = style || frp.signal.constant('#888');

  return {
    bounds: function(view) {
      return rect.value;
    },
    draw: function(view) {
      view.withContext(function(c) {
        var r = rect.value;
        c.save();
        c.fillStyle = style.value;
        c.fillRect(
          r.left,
          r.top,
          r.width,
          r.height
        );
        c.restore();
      });
    }
  };
});


def('image', function(image, origin, clip) {

  origin = origin || image.map(function(image) {
    return [image.width / 2, image.height / 2];
  });

  clip = clip || image.map(function(image) {
    return R.fromSize([0, 0], [image.width, image.height]);
  });

  return {
    bounds: function(view) {
      var sz = [image.value.width, image.value.height];
      return R.fromSize(V.invert(origin.value), sz);
    },
    draw: function(view) {
      view.withContext(function(c) {
        c.drawImage(
          image.value,
          clip.value.left,
          clip.value.top,
          clip.value.width,
          clip.value.height,
          -origin.value[0],
          -origin.value[1],
          clip.value.width,
          clip.value.height
        );
      });
    }
  };
});


def('transform', function(gx, tr) {

  tr = tr || frp.signal.constant(T.identity);

  assert(tr.value.pos);

  return {
    bounds: function(view) {
      return gx.value.bounds(view).transform(tr.value.matrix);
    },
    draw: function(view) {
      gx.value.draw(view.transform(tr.value));
    }
  };
});

exports.zero = {
  bounds: function(view) {
  },
  draw: function(view) {
  }
};

def('fromSignal', function(gx) {
  return {
    bounds: function(view) {
      return gx.value.bounds(view);
    },
    draw: function(view) {
      gx.value.draw(view);
    }
  };
});

def('showHide', function(gx, visible) {

  return {
    bounds: function(view) {
      return gx.value.bounds(view);
    },
    draw: function(view) {
      if (visible.value)
        gx.value.draw(view);
    }
  };
});


def('repeat', function(sGx, sOffset) {

  return {
    draw: function(view) {

      var offset = sOffset ?
        sOffset.value :
        sGx.value.bounds(view).size;

      var startPos = [0, 0];
      var numCopies = [2, 1];

      // project cam bounds to world space
      var vsz = view.getViewport();
      var screenToWorld = view.getTransform().invert();
      var corners = [
        screenToWorld.apply([0,      0]),
        screenToWorld.apply([vsz[0], 0]),
        screenToWorld.apply([vsz[0], vsz[1]]),
        screenToWorld.apply([0,      vsz[1]]),
      ];

      var str =
        corners[0][0]+','+corners[0][1]+' :: '+
        corners[1][0]+','+corners[1][1]+' :: '+
        corners[2][0]+','+corners[2][1]+' :: '+
        corners[3][0]+','+corners[3][1]+' :: ';

      // project again to grid space
      corners = corners.map(function(p) {
        return V.floor(V.div(p, offset));
      });

      // draw all cells in grid within min AABB of projected
      // camera rect
      var xs = util.minMax(_.map(corners, '0'));
      var ys = util.minMax(_.map(corners, '1'));

      var startCoord = [xs.min, ys.min];
      var endCoord = [xs.max + 1, ys.max + 1];

      document.getElementById('dbg').innerHTML = 
        str + '<br>' + 
        corners[0][0]+','+corners[0][1]+' :: '+
        corners[1][0]+','+corners[1][1]+' :: '+
        corners[2][0]+','+corners[2][1]+' :: '+
        corners[3][0]+','+corners[3][1]+' :: '+
        '<br>'+
        startCoord[0]+','+startCoord[1]+'/'+endCoord[0]+','+endCoord[1];

      V.iterect(startCoord, endCoord, function(gridPos) {

        var dPos = V.add(
          startPos,
          V.mul(offset, gridPos)
        );

        var d = T.pos(dPos);
        var w = view.transform(d);

        sGx.value.draw(w);

      });

    }
  };
});


def('tile', function(sGx, doX, doY) {

  doX = doX || frp.signal.constant(true);
  doY = doY || frp.signal.constant(true);

  return {
    draw: function(view) {


      var gx = sGx.value;

      var viewSize = view.getViewport();
      var vtr = view.getTransform();

      var gx = sGx.value;
      var b = gx.bounds(view);
      var gxSize = b.size;
      var vb = new R(
        T.apply(vtr, b.topLeft),
        T.apply(vtr, b.bottomRight)
      );

      var vpos = V.div(vtr.pos, vtr.scale);

      var startPos = V.sub(
        V.sub(
          V.mod(vpos, gxSize),
          vpos
        ),
        gxSize
      );

      var numCopies = V.add(
        V.ceil(V.div(viewSize, gxSize)),
        [1, 1]
      );

      if (!doX.value)
        numCopies[0] = 2;

      if (!doY.value)
        numCopies[1] = 2;

      V.iterect([0, 0], numCopies, function(gridPos) {

        var dPos = V.add(
          startPos,
          V.mul(b.size, gridPos)
        );

        var d = T.pos(dPos);
        var w = view.transform(d);

        gx.draw(w);

      });

    }
  };
});


def('parallax', function(gx, camPos, factor) {

  return {
    draw: function(view) {
      var w = view.transform(T.pos(
        V.mul(
          V.invert(camPos.value),
          V.scaleInverse([1, 1], factor.value)
        )
      ));
      gx.value.draw(w);
    },
    bounds: function(view) {
      return gx.value.bounds(view);
    }
  };
});


def('text', function(text) {

  return {
    draw: function(view) {
      view.withContext(function(c) {
        c.font = '20px Monaco';
        c.textAlign = 'center';
        c.textBaseline = 'middle';
        c.fillStyle = '#fff';
        c.fillText(text.value, 0, 0);
      });
    },
    bounds: function(view) {
    }
  };
});


def('composite', function(gx) {

  var canvas;
  var context;
  var cw, ch;

  return {
    bounds: function(view) { return gx.value.bounds(view); },
    draw: function(view) {

      view.withContext(function(c) {
        var w = c.canvas.width;
        var h = c.canvas.height;
        if (!context || w != cw || h != ch) {
          cw = w;
          ch = h;

          canvas = canvasUtils.makeCanvas({width: w, height: h});
          context = canvas.getContext('2d');

        }
      });

      canvasUtils.clear(context);

      var viewTr = view.getTransform();
      gx.value.draw(makeView(context, viewTr));

      var nv = view.transform(viewTr.invert());
      nv._tr = T.identity;
      nv.withContext(function(c) {
        c.drawImage(canvas, 0, 0);
      });

    }
  };
});


def('alpha', function(gx, alpha) {
  return {
    draw: function(view) {
      gx.value.draw(view.withAlpha(alpha.value));
    },
    bounds: function(view) {
    }
  };
});


def('comp', function(gx, comp) {
  return {
    draw: function(view) {
      gx.value.draw(view.withComp(comp.value));
    },
    bounds: function(view) {
    }
  };
});


def('point', function() {

  var rad = 10;

  return {
    draw: function(view) {
      view.withContext(function(context) {
        context.strokeStyle = '#ff0';
        context.beginPath();
        context.moveTo(-rad, 0);
        context.lineTo( rad, 0);
        context.moveTo(0, -rad);
        context.lineTo(0,  rad);
        context.stroke();
      });
    },
    bounds: function(view) {}
  };
});

def('vector', function(vec) {

  var rad = 10;

  var angle = vec.map(V.norm).map(V.angle);

  var a0 = angle.map(function(a) { return a + 1.1 * Math.PI; });
  var a1 = angle.map(function(a) { return a + 0.9 * Math.PI; });

  var p0s = a0.map(function(a) {
    return V.add(vec.value, V.scale(V.fromAngle(a), rad));
  });

  var p1s = a1.map(function(a) {
    return V.add(vec.value, V.scale(V.fromAngle(a), rad));
  });

  return {
    draw: function(view) {
      view.withContext(function(c) {

        var v = vec.value;
        var p0 = p0s.value;
        var p1 = p1s.value;

        c.strokeStyle = '#ff0';
        c.beginPath();
        c.moveTo(0, 0);
        c.lineTo(v[0], v[1]);
        c.lineTo(p0[0], p0[1]);
        c.moveTo(v[0], v[1]);
        c.lineTo(p1[0], p1[1]);
        c.stroke();
      });
    },
    bounds: function(view) {
    }
  };
});


exports.list = function(xs) {

  var r = {};

  r.items = xs || [];

  if (!r.items.bind)
    r.items = frp.signal.constant(r.items);

  r.bounds = function(view) {};
  r.draw = function(view) {
    _.forEach(r.items.value, function(item) {
      while(item.value)
        item = item.value;
      item.draw(view);
    });
  };

  return r;
};
