var V = require('mxd').Vector;
var _ = require('lodash');
var p2 = require('p2');
var frp = require('frp');
var util = require('ace/util');


exports.airControl = function(ng, pnode, dpad, spec) {
  dpad.bind(function(intent) {
    var x = intent[0];
    var dir = x > 0 ? 1 : -1;

    var spd = pnode.body.velocity[0] * dir;
    var f = (spd < spec.maxSpeed) ? x * spec.force : 0;
    pnode.force = [f, 0];
  });
};


/*
  Cancel all friction with the given footShape if the incline is more
  than 45°
*/
exports.ignoreSteepFriction = function(ng, footShape) {

  var world = ng.physics.world;

  world.on('preSolve', function(e) {

    _.forEach(e.frictionEquations, function(fe) {
      if (fe.shapeA === footShape || fe.shapeB === footShape) {

        var f = fe.shapeB === footShape ? 1 : -1;

        var d = V.dot(fe.t, [-1, 0]);
        if (d * f < .5) {
          fe.enabled = false;
        }
      }
    });

  });
};


/*
  Create a physics body for use in a platform game

  Args:
    ng: Engine
    pos: Position
    envMat: Material of the floor & walls
    walkSpeed: Walk speed
    makeShape: Function returning shape of your body.

  Returns object with keys:
    pnode: Physics node
    grounded: Signal that is true when object is on floor
    setIntent: Call this with -1 or 1 to walk, 0 to stop.
*/
exports.body = function(ng, pos, envMat, walkSpeed, makeShape, offset) {

  var world = ng.physics.world;

  var pnode = ng.physics.node({pos: pos});
  var bodyShape = makeShape();
  var footShape = makeShape();
  pnode.body.fixedRotation = true;
  pnode.body.addShape(bodyShape, offset);
  pnode.body.addShape(footShape, offset);

  var bodyMat = new p2.Material();
  var footMat = new p2.Material();

  bodyShape.material = bodyMat;
  footShape.material = footMat;

  // HEY! foot and body both have restitution applied to all env bodies!
  // Suggests there is a bug! Foot should be floors, body should be walls.

  var footConMat = new p2.ContactMaterial(footMat, envMat, {
    relaxation: 9,
    friction: 1.2,
    restitution: 0
  });
  world.addContactMaterial(footConMat);

  var bodyConMat = new p2.ContactMaterial(bodyMat, envMat, {
    relaxation: 9,
    friction: 0,
    restitution: 0,
  });
  world.addContactMaterial(bodyConMat);

  exports.ignoreSteepFriction(ng, footShape);

  var grounded = nonSteepShapeContacts(ng, pnode, footShape).map(function(x) {
    return x.length != 0;
  });

  return {
    pnode: pnode,
    grounded: grounded,
    setIntent: function(intent) {
      footConMat.surfaceVelocity = intent * walkSpeed;
    },
    remove: function() {
      world.removeContactMaterial(footConMat);
      world.removeContactMaterial(bodyConMat);
      pnode.dead = true;
    }
  };
};

function nonSteepShapeContacts(ng, pnode, shape) {
  return util.shapeContactsFromEvent(ng, nonSteepEvent(pnode), shape);
}

function nonSteepEvent(pnode) {
  return pnode.event.filter(function(e) {
    if (e.type == 'beginContact') {

      if (!e.contactEquations[0])
        return;

      var norm = e.contactEquations[0].normalA;
      var d = V.dot(norm, [0, 1]);
      if (Math.abs(d) < .5)
        return false;
    }
    return true;
  });
}
