var _ = require('lodash');
var frp = require('frp');
var R = require('mxd').Rect;
var V = require('mxd').Vector;
var p2 = require('p2');

var ace = require('ace');
var util = require('ace/util');
var makeLoop = require('ace/loop');
var controls = require('ace/controls');
var T = require('ace/transform');
var graphics = require('ace/graphics');
var makeView = require('ace/view');
var canvasUtils = require('ace/canvasUtils');
var motion = require('ace/motion');

var makePlayer = require('./player');
var assets = require('./assets');
var map = require('./map');
var objects = require('./objects');
var frpList = require('./frpList');
var grow = require('./grow');

var tmxSource = require('../../assets/ld34/map.tmx');

window.onload = function() {
  assets.loaded.bind(function(x) {
    main();
  });
};

function main() {

  // Engine
  var app = ace({
    physics: {gravity: [0, 800]}
  });

  var ng = app.engine;

  var e= document.createElement('div');
  document.body.appendChild(e);
  e.textContent = 'WASD = move, J = jump, K = fire';
  e.style.textAlign='center';
  e.style.fontFamily='sans-serif';


  var m = new p2.ContactMaterial(assets.envMat, assets.envMat, {
    relaxation: 9,
    friction: 1.2,
    restitution: 0
  });
  ng.physics.world.addContactMaterial(m);

  ng.physics.world.solver = new p2.GSSolver();
  ng.physics.world.solver.iterations =  20; // Fast, but contacts might look squishy...

  // Env
  var env = {
    ng: ng,
    markers: {},
    objects: frpList()
  };

  // music
  var music = env.music = document.createElement('audio');
  music.setAttribute('autoplay', 'autoplay');
  music.setAttribute('loop', 'loop');
  music.innerHTML = '<source src="assets/ld34/music.mp3">';
  document.body.appendChild(music);

  // Map
  var tilemap = ace.tilemap.tmx(assets.tiles, tmxSource);
  var tilemapGraphic = tilemap.getGraphic([0, 1]);
  map.addMapCollision(tilemap, ng);

  // Player
  var start = _.find(tilemap.objects, {name: 'start'});
  var player = env.player = makePlayer(ng, start.pos);
  env.objects.push(player);

  // Objects
  tilemap.objects.forEach(function(object) {
    var type = object.object || (object.tile && object.tile.object);
    if (object.pos)
      object.pos  = V.add(object.pos, [32, -16]);
    if (type)
      objects.add(env, type, object);
    else if (object.name)
      env.markers[object.name] = object.pos;
  });

  // Follow player with cam
  var camPos = player.pos.map(_.toArray).map(function(p) {
    return [Math.max(386,Math.min(p[0],3716)), p[1]];
  });

  // Update
  ng.signals.update.watch(function(dt) {
    grow.update(env, dt);
    env.objects.items.value.forEach(function(x) {
      if (x.process)
        x.process(dt);
      if ('ttl' in x) {
        x.ttl -= dt;
        if (x.ttl <= 0) {
          x.die();
        }
      }
    });
  });

  controls.keydown('K').watch(function() {
    objects.add(env, 'growShot', {
      pos: V.add(
        player.pos.value,
        [player.facing.value * 40 - 16, 16]
      ),
      velocity: [player.facing.value * 320, 0]
    });
  });

  var scene = ace.graphics.list([
    tilemapGraphic,
    player.graphic,
    ace.graphics.list(env.objects.map('graphic')),
    graphics.showHide(
      ng.physics.debugGraphic(),
      controls.keydown('P').map(toggler())
    )
  ]);

  scene = ace.graphics.transform(scene, camPos.map(function(camPos) {
    return T.pos(V.scale(camPos,-1));
  }));

  scene = ace.graphics.transform(scene, player.growScale.map(function(x) {
    return T.scale([1/x, 1/x]);
  }));


  scene = ace.graphics.transform(
    scene,
    player.growScale.map(function(x) {
      return T.pos(V.scale([750/2, 500/2], 1));
    })
  );

  var paracam = camPos.map(function(p) {
    return [p[0], 0];
  });

  scene = ace.graphics.list([
    ace.graphics.image(assets.bg0, [0, 0]),

    ace.graphics.parallax(
      ace.graphics.tile(
        ace.graphics.image(assets.bg1, [0, -0]),
        frp.signal.constant(true),
        frp.signal.constant(false)
      ),
      paracam,
      .1
    ),

    ace.graphics.parallax(
      ace.graphics.tile(
        ace.graphics.image(assets.bg2, [0, 0]),
        frp.signal.constant(true),
        frp.signal.constant(false)
      ),
      paracam,
      .2
    ),

    scene
  ]);


  var started = false;
  ng.render.graphic = ace.graphics.image(assets.title, [0, 0]);

  controls.anyKey().filter().watch(function() {
    if (!started) {
      ng.render.graphic = scene;
      started = true;
    }
  });

  /*
  controls.keydown('K').watch(_.partial(throwBomb, env));
  controls.keydown('L').watch(_.partial(sword, env));
  controls.keydown('W').watch(_.partial(enterDoor, env));
  */

}

function toggler() {
  var i = 1;
  return function() {
    i = (i + 1) % 2;
    return !!i;
  };
}
