var _ = require('lodash');
var frp = require('frp');
var assets = require('./assets');
var p2 = require('p2');
var ace = require('ace');
var V = require('mxd').Vector;
var R = require('mxd').Rect;
var T = require('ace/transform');
var grow = require('./grow');
var boss = require('./boss');
var ld34utils = require('./ld34utils');

var types = {};

exports.add = function(env, typeName, spec) {

  var type = types[typeName];

  if (type)
    return type(env, spec);
};

function useGrowSquare(object, smallSize, graphic) {

  var bigSize = V.scale(smallSize, 2);

  var pnode = object.pnode;

  useSmallShape();

  function useSmallShape() {
    removeAllShapes(pnode.body);
    var s = new p2.Box({width:smallSize[0], height:smallSize[1]});
    s.material = assets.envMat;
    pnode.body.addShape(s);
  }

  function useBigShape() {
    removeAllShapes(pnode.body);
    var s = new p2.Box({width:bigSize[0], height:bigSize[1]});
    s.material = assets.envMat;
    pnode.body.addShape(s);
  }

  function use4Shapes() {
    removeAllShapes(pnode.body);
    _.times(4, function() {
      var s = new p2.Box({width:smallSize[0], height:smallSize[1]});
      s.material = assets.envMat;
      pnode.body.addShape(s);
    });
  }

  var scale = frp.cell(1);

  graphic = ace.graphics.transform(
    graphic,
    frp.signal.combine({
      rot: pnode.angle,
      scale: scale
    }).map(function(x) {
      return T.prs([0, 0], x.rot, [x.scale, x.scale]);
    })
  );

  graphic = ace.graphics.list([
    graphic,
    grow.overlayGraphic(object)
  ]);

  graphic = ace.graphics.transform(
    graphic,
    pnode.position.map(T.pos)
  );

  _.assign(object, {
    graphic: graphic,
    growScale: scale,
    growOffsets: [
      [[0, 0], [-16, -16]],
      [[0, 0], [-16,  16]],
      [[0, 0], [ 16, -16]],
      [[0, 0], [ 16,  16]]
    ],
    onGrowStart:   function() { use4Shapes(); },
    onGrowEnd:     function() { useBigShape(); },
    onShrinkStart: function() { use4Shapes(); },
    onShrinkEnd:   function() { useSmallShape(); }
  });

  return scale;
}


function removeAllShapes(body) {
  while (body.shapes.length)
    body.removeShape(body.shapes[0]);
}

types.growShot = function(env, spec) {

  var ng = env.ng;

  var pnode = ng.physics.node({
    pos: spec.pos,
    velocity: spec.velocity
  });

  pnode.body.gravityScale = 0;

  pnode.event.watch(function(e) {
    var target = e.otherNode.object;
    if (target && target.isMirror) {
      if (pnode.body.velocity[0] < 0) {
        pnode.body.velocity[0] *= -1;
      }
      else {
        if (pnode.body.position[0] < target.pnode.body.position[0]) {
          grow(target);
          object.die();
        }
      }
    } else {
      if (target)
        grow(target);
      object.die();
    }
  });

  var s = new p2.Box({width:24, height:16});
  s.material = assets.envMat;
  s.sensor = true;
  pnode.body.addShape(s);

  var graphic = ace.graphics.image(assets.growShot);
  graphic = ace.graphics.transform(graphic, pnode.position.map(function(pos) {
    return T.prs(pos, 0, [pnode.body.velocity[0] > 0 ? 1 : -1, 1]);
  }));

  var object = {
    pnode: pnode,
    graphic: graphic
  };
  pnode.object = object;
  env.objects.push(object);

  object.ttl = 6;

  object.die = function() {
    env.objects.pull(object);
    pnode.dead = true;
  };

  return object;
};

types.box = function(env, spec) {

  var object = {};


  var pnode = env.ng.physics.node({
    pos: V.sub(spec.pos, [16, 16]),
    mass: 100
  });
  object.pnode = pnode;
  pnode.object = object;

  useGrowSquare(object, [32, 32], ld34utils.tileGraphic(spec.tile.id));

  if (spec.lockX) {
    var x = spec.pos[0];
    pnode.body.fixedRotation = true;
    object.process = function() {
      pnode.body.position[0] = x;
    };
  }

  env.objects.push(object);
  return object;
};



types.mirror = function(env, spec) {
  var object = {};

  var pnode = env.ng.physics.node({
    pos: V.sub(spec.pos, [16, 16]),
    mass: 100
  });
  object.pnode = pnode;
  pnode.object = object;

  useGrowSquare(object, [32, 24], ace.graphics.transform(
    ace.graphics.image(assets.mirror),
    T.scale([.5, .5])
  ));

  object.isMirror = true;
  object.growDuration = 4;

  env.objects.push(object);
  return object;
};


types.bomb = function(env, spec) {
  var object = {};

  var pnode = env.ng.physics.node({
    pos: V.sub(spec.pos, [16, 16]),
    mass: 0
  });

  var frame0 = ace.graphics.image(assets.bomb0);
  var frame1 = ace.graphics.image(assets.bomb1);

  var frames = [frame0, frame0, frame0, frame0, frame1];

  var graphic = ace.util.anim(env.ng, frames, .15);

  graphic = ace.graphics.transform(
    graphic,
    pnode.position.map(T.pos)
  );

  object.graphic = graphic;
  env.objects.push(object);
  return object;

};

function trigger(env, pos, w, h, action) {
  var object = {};

  var pnode = env.ng.physics.node({
    pos: V.sub(pos, [16, 16]),
    mass: 0
  });

  var s = new p2.Box({width:w, height:h});
  s.sensor = true;
  pnode.body.addShape(s);

  var done = false;
  pnode.event.watch(function(e) {
    if (done)
      return;
    if (e.otherNode.object && e.otherNode.object.type == 'player') {
      env.objects.pull(object);
      pnode.dead = true;
      done = true;
      action();
    }
  });

  object.graphic = ace.graphics.zero;
  env.objects.push(object);
  return object;
}

types.bossTrigger = function(env, spec) {
  return trigger(env, spec.pos, 16, 200, function() {
    boss.fight(env);
  });
};

types.end = function(env, spec) {
  return trigger(env, spec.pos, 32, 32, function() {
      env.ng.render.graphic = ace.graphics.image(assets.win, [0, 0]);
  });
};
